"""
Base app for analysis apps

"""

from django.db import models
from django.db.models import signals as django_signals
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from installations import models as installations_models
from utils.exchange import exchange
from utils.forms import create_class_with_meta
from utils import reports as utils_reports
from utils.strings import add_suffix_to_string, ugettext_tolanguage as _l
from utils.trees import models as trees_models
from utils.versioning import models as versioning_models
from utils.versioning import versionedmptt

class AppConfig(object):
    # override this

    namespace = "releases"
    tree_root_elementtype = "loc"

    @classmethod
    def get_models(cls):
        return {
##            'suggestionlist': ReleaseSuggestionlist,
##            'scenario': ReleaseScenario,
##            'treeelement': ReleaseTreeElement,
##            'event': ReleaseEvent,
##            'protectionlayer': ReleaseProtectionLayer,
##            'measure': ReleaseMeasure,
##            'protectionlayer_measure': ReleaseProtectionLayer_Measure
##            'attentionpoint': ReleaseAttentionPoint,
        }

    @classmethod
    def get_model(cls, name):
        return cls.get_models().get(name)

    @classmethod
    def get_urls(cls):
        get_urls = __import__('analyses.%s.urls' % cls.namespace, globals(), locals(), ['get_urls',], -1).get_urls
        return get_urls()

class Suggestionlist(versioning_models.VersionedModel):
    name = models.CharField(verbose_name=_('Name'), max_length=100, blank=True, null=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    system = models.BooleanField()

    class Meta:
        verbose_name = _('Suggestion list')
        verbose_name_plural = _('Suggestion lists')
        abstract = True

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""
    appconfig = None

    def __unicode__(self):
        return self.name

    def delete(self, user, *args, **kwargs):
        if not self.version:
            # remove scenarios
            for item in self.scenarios.all():
                item.delete(user, *args, **kwargs)
        super(Suggestionlist, self).delete(user, *args, **kwargs)
    delete.alters_data = True

    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        m = self.appconfig.get_models()
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        # only add self if filterkwarg satisfied:
        if filterkwarg:
            selflist = list(self.__class__.objects.filter(pk=self.pk).filter(**filterkwarg).get_versionedobjects(version=version))
            objects.append(selflist)
        else:
            objects.append([self])

        # scenarios
        scenarios = list(m['scenario'].objects.filter(suggestionlist=self).get_versionedobjects(version=version))

        if scenarios:
            objects.extend(collect_objects_for_scenarios(version, scenarios, filterkwarg, m))
            objects.extend(self.collect_objects_extra(version, scenarios, filterkwarg))

        return objects

    def collect_objects_extra(self, version, scenarios, filterkwarg):
        """
        This should return a list of lists
        """
        return [[]]

    def publish(self, user, instance):
        """
        Create a published version of a suggestion list
        """
        objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

    def dump(self):
        """
        Create a json dump of this suggestion list
        """
        objects = []
        for item in self.collect_objects(filterkwarg={}):
            objects.extend(item)
        return exchange.dump(objects)

    # urls

    @models.permalink
    def get_absolute_url(self, tab='scenarios'):
        return ('%s_suggestionlist' % self.appconfig.namespace, (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='suggestionlist'):
        return ('%s_suggestionlist_edit' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('%s_suggestionlist_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('%s_suggestionlist_report' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        if not version or version == versioning_models.WorkVersion:
            return ('%s_suggestionlist_compare' % self.appconfig.namespace, (self.v_id, 0,))
        return ('%s_suggestionlist_compare' % self.appconfig.namespace, (self.v_id, version.id,))

    @models.permalink
    def get_publish_url(self):
        return ('%s_suggestionlist_publish' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_add_scenario_url(self):
        return ('%s_suggestionlist_add_scenario' % self.appconfig.namespace, (self.v_id,))

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    @classmethod
    def may_add(cls, user):
        return user.has_perm(cls.editpermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.system:
            return False
        return user.has_perm(self.editpermission)

    def may_delete(self, user):
        """
        Only allow delete if the suggestion list has no scenarios
        """
        if self.version:
            return False
        if self.system:
            return False
        if user.has_perm(self.editpermission):
            return self.scenarios.count() == 0
        return False

    def may_publish(self, user):
        """
        Only allow publishing when versioned foreign keys are ok
        """
        if self.version:
            return False
        if self.system:
            return False
        return versioning_models.Version.may_add(user)

    @classmethod
    def get_report_formfields(cls, user):
        from django import forms
        res = [{'name': 'scenarios',
            'field': forms.BooleanField(required=False, label=_('Scenario sheets'),),
            'children': cls.appconfig.get_model('scenario').get_report_formfields(user)}]
        return res

    def get_report_body(self, user, language=u'en', choices={}, level=1):
        from analyses.analysisbase import forms as app_forms
        from analyses.analysisbase import tables as app_tables
        form = app_forms.SuggestionlistSheet(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form,
            'title': u'%s: %s' % (_l('Suggestion list', language), self),
            'level': level,
            }
        scenarios = self.scenarios.all()
        context['scenarios'] = scenarios
        html = [utils_reports.make_pdf_body(context, language=language, template='analysisbase/suggestionlist_report.html')]
        if choices.get('scenarios'):
            for scenario in scenarios:
                html.append('<pdf:nextpage>')
                html.extend(scenario.get_report_body(user, language, choices))
        html.extend(super(Suggestionlist, self).get_report_body(user, language, choices))
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}):
        css = super(Suggestionlist, cls).get_report_css(language=language, choices=choices)
        css.extend(cls.appconfig.get_model('scenario').get_report_css(language, choices))
        return css

    def get_report(self, user, language=u'en', choices={}):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

IS_OK_CHOICES = (
    (0, _('Not evaluated')),
    (1, _('Sufficient')),
    (2, _('Insufficient')),
)

class Scenario(versioning_models.VersionedModel):
    name = models.CharField(verbose_name=_('Name'), max_length=100)
    scope = versioning_models.VersionedForeignKey(installations_models.Scope, verbose_name=_('Scope'), blank=True, null=True, related_name="%(app_label)s_scenario_set")
##    suggestionlist = versioning_models.VersionedForeignKey(DegradationSuggestionlist,
##        verbose_name=_('Suggestion list'), blank=True, null=True,
##        related_name='scenarios')
    type = models.CharField(verbose_name=_('Category'), max_length=100, blank=True, null=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    evaluation = models.TextField(verbose_name = _('Evaluation'), blank=True, null=True)
    is_ok = models.SmallIntegerField(verbose_name=_('Result'), choices=IS_OK_CHOICES, default=0)

    class Meta:
        abstract=True
        verbose_name = _('Scenario')
        verbose_name_plural = _('Scenarios')

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""

    def __unicode__(self):
        return self.name

    # method overrides

    def save(self, user, *args, **kwargs):
        if self.suggestionlist:
            self.scope = None
        super(Scenario, self).save(user, *args, **kwargs)
    save.alters_data = True

    def delete(self, user, *args, **kwargs):
        if not self.version:
            # remove causetreeelements first, in reverse order
            causetreeelements = list(self.get_causetreeelements(version=versioning_models.WorkVersion))
            causetreeelements.reverse()
            for item in causetreeelements:
                item.delete(user, *args, **kwargs)
        super(Scenario, self).delete(user, *args, **kwargs)
    delete.alters_data = True

    def get_causetreeelements(self, version=versioning_models.WorkVersion, select_related=True):
        qs = self.appconfig.get_model('treeelement').tree.filter(scenario=self).get_versionedobjects(version=version)
        if select_related:
            qs = qs.select_related_versioned()
        return qs

    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        m = self.appconfig.get_models()
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        objects.extend(collect_objects_for_scenarios(version, [self], filterkwarg, m))
        objects.extend(self.collect_objects_extra(version, filterkwarg))

        return objects

    def collect_objects_extra(self, version, filterkwarg):
        return []

    def publish(self, user, instance):
        """
        Create a published version of a suggestion list
        """
        objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

    def dump(self):
        """
        Create a json dump of this suggestion list
        """
        objects = []
        for item in self.collect_objects(filterkwarg={}):
            objects.extend(item)
        return exchange.dump(objects)

    # urls

    @models.permalink
    def get_absolute_url(self, tab='causetree'):
        return ('%s_scenario' % self.appconfig.namespace, (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='scenario'):
        return ('%s_scenario_edit' % self.appconfig.namespace, (self.v_id, tab))

    @models.permalink
    def get_delete_url(self):
        return ('%s_scenario_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('%s_scenario_report' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version, tab='scenario'):
        if not version or version == versioning_models.WorkVersion:
            return ('%s_scenario_compare' % self.appconfig.namespace, (self.v_id, 0, tab))
        return ('%s_scenario_compare' % self.appconfig.namespace, (self.v_id, version.id, tab))

    @models.permalink
    def get_move_url(self):
        return ('%s_scenario_move' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_copy_url(self):
        return ('%s_scenario_copy' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_move_to_sl_url(self):
        return ('%s_scenario_move_to_sl' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_copy_to_sl_url(self):
        return ('%s_scenario_copy_to_sl' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_publish_url(self):
        return ('%s_scenario_publish' % self.appconfig.namespace, (self.v_id,))

    @classmethod
    @models.permalink
    def get_add_to_scope_url(cls, scope):
        return ('%s_scope_add_scenario' % cls.appconfig.namespace, (scope.v_id,))

    @classmethod
    @models.permalink
    def get_add_to_scope_via_suggestionlist_url(cls, scope):
        return ('%s_scope_suggestionlists' % cls.appconfig.namespace, (scope.v_id,))

    @classmethod
    def get_type_lookup_url(self):
        return '%s_scenario_type_lookup' % self.appconfig.namespace

    @classmethod
    def get_name_lookup_url(self):
        return '%s_scenario_name_lookup' % self.appconfig.namespace

    # permissions
    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    @classmethod
    def may_add(cls, user, suggestionlist=None, *args, **kwargs):
        if suggestionlist:
            return suggestionlist.may_change(user)
        return user.has_perm(cls.editpermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.suggestionlist:
            return self.suggestionlist.may_change(user)
        else:
            return user.has_perm(self.editpermission)

    def may_delete(self, user):
        """
        also allow delete when scenario has causetree!
        """
        return self.may_change(user)

    def may_publish(self, user):
        """
        Only allow publishing when versioned foreign keys are ok
        """
        if self.version:
            return False
        if versioning_models.Version.may_add(user):
            if self.suggestionlist:
                if self.suggestionlist.__class__.objects.filter(v_id=self.suggestionlist.v_id, published=True).count() == 0:
                    return False
            else:
                if self.scope.__class__.objects.filter(v_id=self.scope.v_id, published=True).count() == 0:
                    return False
            return True
        return False

    def may_copytosl(self, user, *args, **kwargs):
        if self.appconfig.get_model('suggestionlist').objects.filter(system=False).get_versionedobjects(version=versioning_models.WorkVersion).count():
            return user.has_perm(self.appconfig.get_model('suggestionlist').editpermission)
        return False


    # custom methods

    def is_dirty(self):
        """
        Reflect the state of the scenario and the causetree
        """
        if self.version:
            return False
        if self.dirty:
            return True
        if self.appconfig.get_model('treeelement').objects.filter(scenario=self, parent__isnull=True, dirty=True, version__isnull=True).count():
            return True
        return False

    def is_published_dirty(self):
        """
        Reflect the state of the scenario and the causetree
        """
        if self.version:
            return False
        if self.published_dirty:
            return True
        if self.appconfig.get_model('treeelement').objects.filter(scenario=self, parent__isnull=True, published_dirty=True, version__isnull=True).count():
            return True
        return False

    def history(self):
        """
        Return a queryset of all the old saved states of this object.
        Overridden to also return old states of the causetree
        """
        if not hasattr(self, '_history'):
            objectversionids = set(list(self.__class__.objects.filter(v_id=self.v_id, version__isnull=False).values_list('version_id', flat=True).distinct()))
            treeversionids = set(list(self.appconfig.get_model('treeelement').objects.filter(scenario=self, parent__isnull=True, version__isnull=False).values_list('version_id', flat=True).distinct()))
            versionids = list(objectversionids | treeversionids)
            if versionids:
                self._history = versioning_models.Version.objects.filter(id__in=versionids).order_by('-date')
            else:
                self._history = versioning_models.Version.objects.none()
        return self._history

    def history_public(self):
        """
        Return a queryset of all the old saved public states of this object.
        Overridden to also return old states of the causetree
        """
        if not hasattr(self, '_history'):
            objectversionids = set(list(self.__class__.objects.filter(v_id=self.v_id, version__isnull=False, published=True).values_list('version_id', flat=True).distinct()))
            treeversionids = set(list(self.appconfig.get_model('treeelement').objects.filter(scenario=self, parent__isnull=True, version__isnull=False, published=True).values_list('version_id', flat=True).distinct()))
            versionids = list(objectversionids | treeversionids)
            if versionids:
                self._history = versioning_models.Version.objects.filter(id__in=versionids).order_by('-date')
            else:
                self._history = versioning_models.Version.objects.none()
        return self._history

    def copy(self, user, scope=None, suggestionlist=None, suffix=None):
        """
        Copy the scenario and its tree to a scope or suggestionlist
        """
        allowed = False
        if scope:
            allowed = self.__class__.may_add(user, scope=scope)
            kwargs = {'scope': scope, 'suggestionlist': None}
            leafkwargs= {'SL': False}
        elif suggestionlist:
            allowed = self.__class__.may_add(user, suggestionlist=suggestionlist)
            kwargs = {'scope': None, 'suggestionlist': suggestionlist, 'is_ok': False, 'evaluation': u''}
            leafkwargs= {'SL': True}

        if allowed:
            treeobjects = list(self.get_causetreeelements(version=self._version, select_related=True))
            newscen = super(Scenario, self).copy(user, save=False, **kwargs)
            if suffix:
                newscen.name = add_suffix_to_string(self.name, suffix, 100)
            newscen.save(user)
            newelements = self.appconfig.get_model('treeelement').copy_tree(treeobjects, user, scenario=newscen, leafkwargs=leafkwargs)
            self.copy_extra(user, newelements)
            return newscen

    def copy_extra(self, user, newelements):
        pass

    def move(self, user, scope=None, suggestionlist=None):
        """
        Move the scenario and its tree to a scope or suggestionlist
        """
        odsl = bool(self.suggestionlist)
        if scope and scope != self.scope:
            self.scope = scope
            self.suggestionlist = None
            self.save(user)
        if suggestionlist and suggestionlist != self.suggestionlist:
            self.scope = None
            self.suggestionlist = suggestionlist
            self.evaluation = u''
            self.is_ok = False
            self.save(user)
        # change SL boolean on protectionlayers if needed
        if bool(self.suggestionlist) != odsl:
            version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
            treeobjects = self.appconfig.get_model('treeelement').tree.filter(scenario=self
                ).get_versionedobjects(version=version)
            c = treeobjects.filter(elementtype = PROTECTIONLAYER_ID).values_list('object_id', flat=True)
            if c:
                pls = self.appconfig.get_models()['protectionlayer'].objects.filter(v_id__in=c).get_versionedobjects(version=version)
                for item in pls:
                    item.SL = bool(self.suggestionlist)
                    item.save(user)

    def add_to_extra_context(self, extra_context, object):
        if self.render_pdf:
            treeobjects = self.appconfig.get_model('treeelement').tree.filter(scenario=object
                ).get_versionedobjects(version=self.version).select_related_versioned()
            extra_context['treeobjects'] = trees_views.annotate_tree(treeobjects)
            extra_context['tree_details'] = self.choices.get('tree_details', False)

    @classmethod
    def get_listreport_formfields(cls, user):
        from django import forms
        res = [{'name': 'scenarios',
            'field': forms.BooleanField(required=False, label=_('Scenario sheets'),),
            'children': cls.get_report_formfields(user)}]
        return res

    @classmethod
    def get_report_formfields(cls, user):
        from django import forms
        res = [{'name': 'causes',
                'field': forms.BooleanField(required=False, label=_('Cause details'),),
                'children': cls.appconfig.get_model('event').get_report_formfields(user)},
            {'name': 'protectionlayers',
                'field': forms.BooleanField(required=False, label=_('Protection layer details'),),
                'children': cls.appconfig.get_model('protectionlayer').get_report_formfields(user)},
            ]
        return res

    def get_report_body(self, user, language=u'en', choices={}, level=1):
        from analyses.analysisbase import forms as app_forms
        from utils.trees import views as trees_views

        form = create_class_with_meta(app_forms.ScenarioReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form, 'title': u'%s: %s' % (_l('Scenario', language), self), 'level': level}
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        treeobjects = self.appconfig.get_model('treeelement').tree.filter(scenario=self
            ).get_versionedobjects(version=version).select_related_versioned()
        context['treeobjects'] = trees_views.annotate_tree(treeobjects)
        html = [utils_reports.make_pdf_body(context, language=language, template='analysisbase/scenario_report.html')]
        html.extend(super(Scenario, self).get_report_body(user, language, choices, level))
        if choices.get('causes'):
            # include cause details
            c = list(treeobjects.filter(elementtype = EVENT_ID).values_list('object_id', flat=True))
            if c:
                causes = self.appconfig.get_models()['event'].objects.filter(v_id__in=c).get_versionedobjects(version=version)
                causehtml = []
                for item in causes:
                    causehtml.extend(item.get_report_body(user, language, choices, level+2))
                if causehtml:
                    html.append(u'<p class="header%s">%s</p>\n' % (level+1, _l('Cause details', language)))
                    html.extend(causehtml)
        if choices.get('protectionlayers'):
            # include protectionlayer details
            c = list(treeobjects.filter(elementtype = PROTECTIONLAYER_ID).values_list('object_id', flat=True))
            if c:
                pls = self.appconfig.get_models()['protectionlayer'].objects.filter(v_id__in=c).get_versionedobjects(version=version)
                plshtml = []
                for item in pls:
                    plshtml.extend(item.get_report_body(user, language, choices, level+2))
                if plshtml:
                    html.append(u'<p class="header%s">%s</p>\n' % (level+1, _l('Protection layer details', language)))
                    html.extend(plshtml)
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}):
        css = super(Scenario, cls).get_report_css(language=language, choices=choices)
        css.append('analysisbase/scenario_report.css')
        if choices.get('causes'):
            css.extend(cls.appconfig.get_model('event').get_report_css(language=language, choices=choices))
        if choices.get('protectionlayers'):
            css.extend(cls.appconfig.get_model('protectionlayer').get_report_css(language=language, choices=choices))
        return css

    def get_report(self, user, language=u'en', choices={}):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

EVENT_ID = 1
PROTECTIONLAYER_ID = 2

class TreeElement(trees_models.TreeElement):
##    degradationscenario = versioning_models.VersionedForeignKey(DegradationScenario, verbose_name = _('Scenario'))
    # false = OR, true = AND
    gate = models.BooleanField()

    class Meta:
        abstract=True

    TreeElementChoices = (
        (EVENT_ID, _('Event')),
        (PROTECTIONLAYER_ID, _('Protection layer')),
    )

    # permissions

    @classmethod
    def may_add(cls, user, parent):
        return parent.may_change(user)

    # custom methods

    @classmethod
    def collect_items(cls, version):
        return list(('Causetree for scenario', item.scenario, None) for item in cls.objects.filter(version=version, parent__isnull=True, deleted=False))

    def get_elementtype_for_tree(self):
        if self.is_root_node():
            return self.appconfig.tree_root_elementtype
        return self._type_fk().get_elementtype_for_tree()

class Event(trees_models.TreeLeaf):

    class Meta:
        abstract=True
        verbose_name = _('Event')
        verbose_name_plural = _('Event')

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""

    # method overrides
    # from Treeleaf:
    def get_elementtype_for_tree(self):
        return 'event'

    def get_tree_class(self):
        return self.appconfig.get_model('treeelement')

    def create_treeelement(self, **kwargs):
        scenario = kwargs.pop('scenario', None)
        if not scenario:
            raise Exception("create_treeelement called on event in %s app without scenario" % self.appconfig.namespace)
        return self.appconfig.get_model('treeelement')(scenario=scenario, elementtype=EVENT_ID)

    # urls

    @models.permalink
    def get_absolute_url(self, tab=None):
        return ('%s_event' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_edit_url(self, tab=None):
        return ('%s_event_edit' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('%s_event_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('%s_event_compare' % self.appconfig.namespace, (self.v_id, version.id))

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.get_treeelements().count() == 0:
            return user.has_perm(self.editpermission)
        for item in self.get_treeelements().select_related_versioned():
            if not item.scenario.may_change(user):
                return False
        return True

    def may_delete(self, user):
        if self.version:
            return False
        if self.get_treeelements().count() == 0:
            return user.has_perm(self.editpermission)
        for item in self.get_treeelements().select_related_versioned():
            if not item.parent:
                return False
            if item.get_children():
                return False
            if not item.scenario.may_change(user):
                return False
        return True

    @classmethod
    def get_report_formfields(cls, user):
        return []

    def get_report_reportsheet(self):
        from analyses.analysisbase import forms as app_forms
        return create_class_with_meta(app_forms.EventSheet, model=self.__class__)

    def has_report_form(self):
        if self.description and len(self.description):
            return True
        return False

    def get_report_body(self, user, language=u'en', choices={}, level=1, **kwargs):
        subhtml = super(Event, self).get_report_body(user, language, choices, level)
        if level>1:
            if not subhtml:
                if not self.has_report_form():
                    return []
        form = self.get_report_reportsheet()(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form,
            'title': u'%s: %s' % (_l('Cause', language), self),
            'level': level}
        html = [utils_reports.make_pdf_body(context, language=language, template='versioning/report.html')]
        html.extend(subhtml)
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}, **kwargs):
        css = super(Event, cls).get_report_css(language=language, choices=choices)
        return css

    def get_report(self, user, language=u'en', choices={}, **kwargs):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

class ProtectionLayer(trees_models.TreeLeaf):
    """Used for both control functions as safety functions
    """
    effectivity = models.TextField(verbose_name = _('Effectivity'), blank=True, null=True)
    reliability = models.TextField(verbose_name = _('Reliability'), blank=True, null=True)
    risks = models.TextField(verbose_name = _('Risks'), blank=True, null=True)
    # Denorm field
    SL = models.BooleanField(verbose_name=_('Suggestion list'))

    class Meta:
        abstract=True
        verbose_name = _('Protection layer')
        verbose_name_plural = _('Protection layers')

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""

    def get_tree_class(self):
        return self.appconfig.get_model('treeelement')

    def create_treeelement(self, **kwargs):
        scenario = kwargs.pop('scenario', None)
        if not scenario:
            raise Exception("create_treeelement called on protectionlayer in %s app without scenario" % self.appconfig.namespace)
        return self.appconfig.get_model('treeelement')(scenario=scenario, elementtype=PROTECTIONLAYER_ID)

    def get_elementtype_for_tree(self):
        return 'protectionlayer'

    def delete(self, user, *args, **kwargs):
        # delete pmes
        for item in self.appconfig.get_model('protectionlayer_measure').objects.filter(protectionlayer=self).get_versionedobjects(version=versioning_models.WorkVersion):
            item.delete(user)
        return super(ProtectionLayer, self).delete(user, *args, **kwargs)
    delete.alters_data = True

    # Note about copy:
    # no copy override: we do not copy the measure nor the PL_measure

    # urls

    @models.permalink
    def get_absolute_url(self, tab=None):
        return ('%s_protectionlayer' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_edit_url(self, tab=None):
        return ('%s_protectionlayer_edit' % self.appconfig.namespace, (self.v_id, ))

    @models.permalink
    def get_delete_url(self):
        return ('%s_protectionlayer_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('%s_protectionlayer_compare' % self.appconfig.namespace, (self.v_id, version.id))

    @models.permalink
    def get_add_measure_url(self):
        return ('%s_protectionlayer_add_measure' % self.appconfig.namespace, (self.v_id,))

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.get_treeelements().count() == 0:
            return user.has_perm(self.editpermission)
        for item in self.get_treeelements().select_related_versioned():
            if not item.scenario.may_change(user):
                return False
        return True

    def may_delete(self, user):
        if self.version:
            return False
        if self.get_treeelements().count() == 0:
            return user.has_perm(self.editpermission)
        for item in self.get_treeelements().select_related_versioned():
            if not item.parent:
                return False
            if item.get_children():
                return False
            if not item.scenario.may_change(user):
                return False
        return True

    @classmethod
    def get_report_formfields(cls, user):
        from django import forms
        res = [{'name': 'measures',
                'field': forms.BooleanField(required=False, label=_('Measure details'),),
                'children': cls.appconfig.get_model('measure').get_report_formfields(user)}]
        return res

    def get_report_reportsheet(self):
        from analyses.analysisbase import forms as app_forms
        return create_class_with_meta(app_forms.ProtectionLayerSheet, model=self.__class__)

    def has_report_form(self):
        if self.description and len(self.description):
            return True
        if self.effectivity and len(self.effectivity):
            return True
        if self.reliability and len(self.reliability):
            return True
        if self.risks and len(self.risks):
            return True
        return False

    def get_report_body(self, user, language=u'en', choices={}, level=1, **kwargs):
        from analyses.analysisbase import tables as app_tables

        measures = None
        measurecount = 0
        # create subreports first, so we know whether there are measures
        meshtml = []
        plmes = self.appconfig.get_model('protectionlayer_measure').objects.filter(protectionlayer=self).get_versionedobjects(version=self._version).values_list('measure', flat=True).distinct()
        if plmes:
            measures = self.appconfig.get_model('measure').objects.filter(v_id__in=plmes).get_versionedobjects(version=self._version)
            measurecount = measures.count()
        if choices.get('measures') and measurecount:
            for item in measures:
                meshtml.extend(item.get_report_body(user, language, choices, level+2))

        subhtml = super(ProtectionLayer, self).get_report_body(user, language, choices, level)

        # if no measures or no description: bail out
        if level>1:
            if not measurecount and not meshtml and not subhtml:
                if not self.has_report_form():
                    return []

        form = self.get_report_reportsheet()(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form,
            'title': u'%s: %s' % (_l('Protection layer', language), self),
            'level': level}
        tables = []
        if measurecount and not choices.get('measures'):
            table = create_class_with_meta(app_tables.MeasuresTable, model=self.appconfig.get_model('measure'))(measures)
            table.title = _l('Measures', language)
            table.classes = ['measures']
            tables.append(table)
        context['tables'] = tables
        html = [utils_reports.make_pdf_body(context, language=language, template='versioning/report.html'),]
        if meshtml:
            html.append(u'<p class="header%s">%s</p>\n' % (level+1, _l('Measures', language)))
            html.extend(meshtml)
        html.extend(subhtml)
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}, **kwargs):
        css = super(ProtectionLayer, cls).get_report_css(language=language, choices=choices)
        css.extend(cls.appconfig.get_model('measure').get_listreport_css(language=language, choices=choices))
        if choices.get('measures'):
            css.extend(cls.appconfig.get_model('measure').get_report_css(language=language, choices=choices))
        return css

    def get_report(self, user, language=u'en', choices={}, **kwargs):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

class Measure(versioning_models.VersionedModel):
    """Used for implementation of protection layer
    """
    name = models.CharField(verbose_name=_('Name'), max_length=100)
    tag = models.CharField(verbose_name=_('Tag'), max_length=100, blank=True, null=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    type = models.CharField(verbose_name=_('Category'), max_length=100, blank=True, null=True, db_index=True)
    SL = models.BooleanField(verbose_name=_('Suggestion list'))

    class Meta:
        abstract=True
        verbose_name = _('Measure')
        verbose_name_plural = _('Measures')

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""

    def __unicode__(self):
        return self.name

# TODO: make sure SL boolean gets set correctly ??

    # method overrides
    def delete(self, user, *args, **kwargs):
        # delete pmes
        for item in self.appconfig.get_model('protectionlayer_measure').objects.filter(measure=self).get_versionedobjects(version=versioning_models.WorkVersion):
            item.delete(user)
        # delete attentionpoints
        for ap in self.attentionpoints.all():
            ap.delete(user)
        return super(Measure, self).delete(user, *args, **kwargs)
    delete.alters_data = True

    # from VersionedModel
    def copy(self, user, suffix=None, **kwargs):
        objectlist = list(self.attentionpoints.all())
        newpl = super(Measure, self).copy(user, **kwargs)
        if suffix:
            newpl.name = add_suffix_to_string(self.name, suffix, 100)
        newpl.save(user)
        for item in objectlist:
            item.copy(user, measure=newpl)
        return newpl

    # urls

    @models.permalink
    def get_absolute_url(self, tab='description'):
        return ('%s_measure' % self.appconfig.namespace, (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='description'):
        return ('%s_measure_edit' % self.appconfig.namespace, (self.v_id, ))

    @models.permalink
    def get_delete_url(self):
        return ('%s_measure_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('%s_measure_report' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('%s_measure_compare' % self.appconfig.namespace, (self.v_id, version.id))

    @models.permalink
    def get_replace_url(self):
        return ('%s_measure_replace' % self.appconfig.namespace , (self.v_id,))

    @models.permalink
    def get_copy_url(self):
        return ('%s_measure_copy' % self.appconfig.namespace , (self.v_id,))

    @models.permalink
    def get_add_attentionpoint_url(self):
        return ('%s_measure_add_attentionpoint' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_add_attentionpoint_from_suggestionlist_url(self):
        return ('%s_measure_apsuggestionlists' % self.appconfig.namespace, (self.v_id,))

    @classmethod
    def get_name_lookup_url(self):
        return '%s_measure_name_lookup' % self.appconfig.namespace

    @classmethod
    def get_type_lookup_url(self):
        return '%s_measure_type_lookup' % self.appconfig.namespace

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    @classmethod
    def may_add(cls, user, suggestionlist=None, *args, **kwargs):
        if suggestionlist:
            return suggestionlist.may_change(user)
        return user.has_perm(cls.editpermission)

    def may_change(self, user):
        if self.version:
            return False
        return user.has_perm(self.editpermission)

    def may_delete(self, user):
        if self.version:
            return False
# allow deleting measures with attention points
# allow deleting measures that are linked
#        if self.attentionpoints.count() or self.protectionlayers.count():
#            return False
        return user.has_perm(self.editpermission)

    def may_replace(self, user):
        if self.version:
            return False
# allow replacing measures with attention points
#        if self.attentionpoints.count():
#            return False
        return user.has_perm(self.editpermission)


    @classmethod
    def get_listreport_formfields(cls, user):
        from django import forms
        res = [{'name': 'measures',
            'field': forms.BooleanField(required=False, label=_('Measure sheets'),),
            'children': cls.get_report_formfields(user)}]
        return res

    @classmethod
    def get_listreport_css(cls, language=u'en', choices={}, **kwargs):
        return ['analysisbase/measures_report.css']

    @classmethod
    def get_report_formfields(cls, user):
        from django import forms
        return [{'name': 'attentionpoints',
            'field': forms.BooleanField(required=False, label=_('Attention point details'),),
            'children': []}]

    def get_report_body(self, user, language=u'en', choices={}, level=1, **kwargs):
        from analyses.analysisbase import tables as app_tables
        from analyses.analysisbase import forms as app_forms

        ap = self.attentionpoints.all()
        # create subreports first, if they are empty and this sheet has no extra info, ignore it
        aphtml = []
        if choices.get('attentionpoints'):
            for attentionpoint in ap:
                aphtml.extend(attentionpoint.get_report_body(user, language, choices, level+2, for_pl=True))
        subhtml = super(Measure, self).get_report_body(user, language, choices, level)

        form = create_class_with_meta(app_forms.MeasureReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form,
            'title': u'%s: %s' % (_l('Measure', language), self),
            'level': level}
        tables = []
        if ap and not choices.get('attentionpoints'):
            table = create_class_with_meta(app_tables.AttentionPointsTable, model=self.appconfig.get_models()['attentionpoint'])(ap)
            table.title = _l('Attention points', language)
            table.classes = ['attentionpoints']
            tables.append(table)
        if level==1:
            queryset = self.protectionlayers.all()
            table = create_class_with_meta(app_tables.MeasureUsageTable, model=self.appconfig.get_models()['protectionlayer'])(queryset)
            table.title = _l('Usage', language)
            tables.append(table)
        context['tables'] = tables
        html = [utils_reports.make_pdf_body(context, language=language, template='versioning/report.html')]
        if aphtml:
            html.append(u'<p class="header%s">%s</p>\n' % (level+1, _l('Attention points', language)))
            html.extend(aphtml)
        html.extend(subhtml)
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}, **kwargs):
        css = super(Measure, cls).get_report_css(language=language, choices=choices)
        css.append('analysisbase/measure_report.css')
        if choices.get('attentionpoints'):
            css.extend(cls.appconfig.get_model('attentionpoint').get_report_css(language=language, choices=choices))
        return css

    def get_report(self, user, language=u'en', choices={}, **kwargs):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

# Explicit m2m model to allow versioning
class ProtectionLayer_Measure(versioning_models.VersionedModel):
##    protectionlayer = versioning_models.VersionedForeignKey(ProtectionLayer)
##    measure = versioning_models.VersionedForeignKey(Measure)

    class Meta:
        abstract=True
        verbose_name = _('Measure for protection layer')
        verbose_name_plural = _('Measures for protection layer')

    def __unicode__(self):
        return u'%s for %s' % (self.measure, self.protectionlayer)

    @models.permalink
    def get_delete_url(self):
        return ('%s_protectionlayer_measure_delete' % self.appconfig.namespace, (self.v_id,))

    def may_delete(self, user):
        if self.version:
            return False
        return user.has_perm(self.editpermission)

    @classmethod
    def collect_items(cls, version):
        modified = list(cls.objects.filter(version=version).values_list('protectionlayer_id', flat=True).distinct())
        if modified:
            items = cls.appconfig.get_model('protectionlayer').objects.filter(v_id__in=modified).get_versionedobjects(version=version)
            return list(('Measures for protectionlayer', item, None) for item in items)
        return []

class APSuggestionlist(versioning_models.VersionedModel):
    name = models.CharField(verbose_name=_('Name'), max_length=100, blank=True, null=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    system = models.BooleanField()

    class Meta:
        abstract = True
        verbose_name = _('Attention point suggestion list')
        verbose_name_plural = _('Attention point suggestion lists')

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""
    appconfig = None

    def __unicode__(self):
        return self.name

    def delete(self, user, *args, **kwargs):
        if not self.version:
            # remove attentionpoints
            for item in self.attentionpoints.all():
                item.delete(user, *args, **kwargs)
        super(APSuggestionlist, self).delete(user, *args, **kwargs)
    delete.alters_data = True

    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        m = self.appconfig.get_models()
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        # only add self if filterkwarg satisfied:
        if filterkwarg:
            selflist = list(self.__class__.objects.filter(pk=self.pk).filter(**filterkwarg).get_versionedobjects(version=version))
            objects.append(selflist)
        else:
            objects.append([self])

        # attention points
        objects.append(list(m['attentionpoint'].objects.filter(**filterkwarg).filter(suggestionlist=self).get_versionedobjects(version=version)))

        return objects

    def publish(self, user, instance):
        """
        Create a published version of a suggestion list
        """
        objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

    def dump(self):
        """
        Create a json dump of this suggestion list
        """
        objects = []
        for item in self.collect_objects(filterkwarg={}):
            objects.extend(item)
        return exchange.dump(objects)

    # urls

    @models.permalink
    def get_absolute_url(self, tab='attentionpoints'):
        return ('%s_apsuggestionlist' % self.appconfig.namespace, (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='suggestionlist'):
        return ('%s_apsuggestionlist_edit' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('%s_apsuggestionlist_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('%s_apsuggestionlist_report' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        if not version or version == versioning_models.WorkVersion:
            return ('%s_apsuggestionlist_compare' % self.appconfig.namespace, (self.v_id, 0,))
        return ('%s_apsuggestionlist_compare' % self.appconfig.namespace, (self.v_id, version.id,))

    @models.permalink
    def get_publish_url(self):
        return ('%s_apsuggestionlist_publish' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_add_attentionpoint_url(self):
        return ('%s_apsuggestionlist_add_attentionpoint' % self.appconfig.namespace, (self.v_id,))

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    @classmethod
    def may_add(cls, user):
        return user.has_perm(cls.editpermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.system:
            return False
        return user.has_perm(self.editpermission)

    def may_delete(self, user):
        """
        Only allow delete if the suggestion list has no scenarios
        """
        if self.version:
            return False
        if self.system:
            return False
        if user.has_perm(self.editpermission):
            return self.attentionpoints.count() == 0
        return False

    def may_publish(self, user):
        """
        Only allow publishing when versioned foreign keys are ok
        """
        if self.version:
            return False
        if self.system:
            return False
        return versioning_models.Version.may_add(user)

    @classmethod
    def get_report_formfields(cls, user):
        return []

    def get_report_body(self, user, language=u'en', choices={}):
        from analyses.analysisbase import forms as app_forms
        form = create_class_with_meta(app_forms.ApSuggestionlistReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
        title = u'%s: %s' % (_l('Attention point suggestion list', language), self)
        context = {'object': self, 'form': form,
            'title': title,
            'level': 1,
        }
        html = [utils_reports.make_pdf_body(context, language=language, template='analysisbase/apsuggestionlist_report.html')]
        html.extend(super(APSuggestionlist, self).get_report_body(user, language, choices))
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}, **kwargs):
        css = super(APSuggestionlist, cls).get_report_css(language=language, choices=choices)
        css.append('analysisbase/apsuggestionlist_report.css')
        return css

    def get_report(self, user, language=u'en', choices={}, **kwargs):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

class AttentionPoint(versioning_models.VersionedModel):
    """
    Attention point base model
    Subclasses VersionedModel from versioning app
    Subclasses should define a measure field pointing to a measure model
    """
##    measure = versioning_models.VersionedForeignKey(DegradationMeasure, verbose_name = _('Measure'), blank=True, null=True, related_name='attentionpoints')
##    suggestionlist = versioning_models.VersionedForeignKey(APSuggestionlist, verbose_name=_('Suggestion list'), blank=True, null=True, related_name='attentionpoints')
    name = models.CharField(verbose_name=_('Name'), max_length=100)
    type = models.CharField(verbose_name=_('Category'), max_length=100, blank=True, null=True, db_index=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    # This is actually a denormalisation since pl is already marked
    SL = models.BooleanField(verbose_name=_('Suggestion list'))

    class Meta:
        abstract = True
        verbose_name = _('Attention point')
        verbose_name_plural = _('Attention points')
        ordering = ('type', 'name',)

    # make sure to override these:
    displaypermisssion = ""
    editpermission = ""

    def __unicode__(self):
        return self.name

    # method overrides

    def save(self, user, *args, **kwargs):
        """
        Doing the denormalisation for SL
        """
        if self.suggestionlist:
            self.measure = None
        if self.measure:
            self.SL = self.measure.SL
        super(AttentionPoint, self).save(user, *args, **kwargs)

    # urls

    @models.permalink
    def get_absolute_url(self, tab=None):
        return ('%s_attentionpoint' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_edit_url(self, tab=None):
        return ('%s_attentionpoint_edit' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('%s_attentionpoint_delete' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('%s_attentionpoint_report' % self.appconfig.namespace, (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('%s_attentionpoint_compare' % self.appconfig.namespace, (self.v_id, version.id))

    @classmethod
    def get_name_lookup_url(self):
        return '%s_attentionpoint_name_lookup' % self.appconfig.namespace

    @classmethod
    def get_type_lookup_url(self):
        return '%s_attentionpoint_type_lookup' % self.appconfig.namespace

    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm(cls.displaypermission)

    @classmethod
    def may_add(cls, user, parent=None, suggestionlist=None, *args, **kwargs):
        if suggestionlist:
            return suggestionlist.may_change(user)
        if parent:
            return parent.may_change(user)
        return user.has_perm(cls.editpermission)

    def may_change(self, user):
        if self.version:
            return False
        if self.suggestionlist:
            return self.suggestionlist.may_change(user)
        elif self.measure:
            return self.measure.may_change(user)
        return False

    def may_delete(self, user):
        return self.may_change(user)

    def copy(self, user, measure=None, suggestionlist=None, suffix=None):
        """
        Copy the attention point to a measure or suggestionlist
        """
        allowed = False
        if measure:
            allowed = self.__class__.may_add(user, parent=measure)
            kwargs = {'measure': measure, 'suggestionlist': None}
        elif suggestionlist:
            allowed = self.__class__.may_add(user, suggestionlist=suggestionlist)
            kwargs = {'protectionlayer': None, 'suggestionlist': suggestionlist}
        if allowed:
            newap = super(AttentionPoint, self).copy(user, save=False, **kwargs)
            if suffix:
                newap.name = add_suffix_to_string(self.name, suffix, 100)
            newap.save(user)
            return newap

    def move(self, user, measure=None, suggestionlist=None):
        """
        Move the attention point to a measure or suggestionlist
        """
        if measure and measure != self.measure:
            self.measure = measure
            self.suggestionlist = None
            self.save(user)
        if suggestionlist and suggestionlist != self.suggestionlist:
            self.measure = None
            self.suggestionlist = suggestionlist
            self.save(user)

    @classmethod
    def get_listreport_formfields(cls, user):
        from django import forms
        res = [{'name': 'attentionpoints',
            'field': forms.BooleanField(required=False, label=_('Attention point details'),),
            'children': cls.get_report_formfields(user)}]
        return res

    @classmethod
    def get_listreport_css(cls, language=u'en', choices={}, **kwargs):
        return ['analysisbase/attentionpoints_report.css']

    @classmethod
    def get_report_formfields(cls, user):
        return []

    def get_report_body(self, user, language=u'en', choices={}, level=1, for_pl=False, **kwargs):
        from analyses.analysisbase import forms as app_forms
        subhtml = super(AttentionPoint, self).get_report_body(user, language, choices, level)
        if level>1:
            if for_pl:
                form = create_class_with_meta(app_forms.AttentionPointSubReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
            else:
                form = create_class_with_meta(app_forms.AttentionPointReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
        else:
            form = create_class_with_meta(app_forms.AttentionPointReportSheet, model=self.__class__)(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form,
            'title': u'%s: %s' % (_l('Attention point', language), self),
            'level': level}
        html = [utils_reports.make_pdf_body(context, language=language, template='versioning/report.html')]
        html.extend(subhtml)
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}, **kwargs):
        css = super(AttentionPoint, cls).get_report_css(language=language, choices=choices)
        return css

    def get_report(self, user, language=u'en', choices={}, **kwargs):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

class AllSuggestionlists(object):
    """
    Dummy object for creating reports with all suggestionlists
    """

    @classmethod
    def get_report_formfields(cls, appconfig, user):
        from django import forms
        res = []
        if appconfig.get_model('suggestionlist').may_display(user):
            res.append({'name': 'suggestionlists',
                'field': forms.BooleanField(required=False, label=_('Scenario suggestion lists'),),
                'children': appconfig.get_model('suggestionlist').get_report_formfields(user)})
        if appconfig.get_model('apsuggestionlist').may_display(user):
            res.append({'name': 'apsuggestionlists',
                'field': forms.BooleanField(required=False, label=_('Attention point suggestion lists'),),
                'children': appconfig.get_model('apsuggestionlist').get_report_formfields(user)})
        return res

    @classmethod
    def get_report_body(cls, appconfig, user, language=u'en', choices={}, version=versioning_models.WorkVersion, level=1):
        from analyses.analysisbase import tables as app_tables
        if appconfig.get_model('suggestionlist').may_display(user):
            suggestionlists = appconfig.get_model('suggestionlist').objects.all().get_versionedobjects(version)
        else:
            suggestionlists = []
        if appconfig.get_model('suggestionlist').may_display(user):
            apsuggestionlists = appconfig.get_model('apsuggestionlist').objects.all().get_versionedobjects(version)
        else:
            apsuggestionlists = []

        context = {'suggestionlists': suggestionlists, 'apsuggestionlists': apsuggestionlists,
            'title': u'%s %s' % (_l('Suggestion lists', language), _l(appconfig.analysistype, language)),
            'level': level}
        tables = []
        if suggestionlists:
            table = create_class_with_meta(app_tables.SuggestionListsTable, model=appconfig.get_models()['suggestionlist'])(suggestionlists)
            table.title = _l('Scenario suggestion lists', language)
            tables.append(table)
        if apsuggestionlists:
            table = create_class_with_meta(app_tables.SuggestionListsTable, model=appconfig.get_models()['apsuggestionlist'])(apsuggestionlists)
            table.title = _l('Attention point suggestion lists', language)
            tables.append(table)
        context['tables'] = tables

        html = [utils_reports.make_pdf_body(context, language=language, template='versioning/report.html')]
        if choices.get('suggestionlists'):
            for suggestionlist in suggestionlists:
                html.append('<pdf:nextpage>')
                html.extend(suggestionlist.get_report_body(user, language, choices))
        if choices.get('apsuggestionlists'):
            for apsuggestionlist in apsuggestionlists:
                html.append('<pdf:nextpage>')
                html.extend(apsuggestionlist.get_report_body(user, language, choices))
        return html

    @classmethod
    def get_report_css(cls, appconfig, language=u'en', choices={}):
        css = []
        css.extend(appconfig.get_model('suggestionlist').get_report_css(language, choices))
        css.extend(appconfig.get_model('apsuggestionlist').get_report_css(language, choices))
        return css

    @classmethod
    def get_report(cls, appconfig, user, language=u'en', choices={}, version=versioning_models.WorkVersion):
        body_strings = cls.get_report_body(appconfig, user, language, choices, version)
        css_files = cls.get_report_css(appconfig, language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i+n]

def collect_objects_for_scenarios(version, scenarios, filterkwarg, models):
    objects = []
    if scenarios:
        sc = [item.v_id for item in scenarios]
        for sc_chunk in chunks(sc, 100):
            objects.append(list(models['scenario'].objects.filter(**filterkwarg).filter(v_id__in=sc_chunk).get_versionedobjects(version=version)))

            # causes
            items = models['treeelement'].objects.filter(scenario__in=sc_chunk
                ).filter(elementtype=EVENT_ID
                ).get_versionedobjects(version=version
                ).values_list('object_id', flat=True)
            if items:
                for items_chunk in chunks(items, 100):
                    objects.append(list(models['event'].objects.filter(**filterkwarg).filter(v_id__in=items_chunk).get_versionedobjects(version=version)))

            # protectionlayers
            items = models['treeelement'].objects.filter(scenario__in=sc_chunk
                ).filter(elementtype=PROTECTIONLAYER_ID
                ).get_versionedobjects(version=version
                ).values_list('object_id', flat=True)
            if items:
                for items_chunk in chunks(items, 100):
                    protectionlayers = list(models['protectionlayer'].objects.filter(**filterkwarg).filter(v_id__in=items_chunk).get_versionedobjects(version=version))
                    objects.append(protectionlayers)
                    plids = [pl.v_id for pl in protectionlayers]
                    if plids:
                        for plids_chunk in chunks(plids, 100):
                            plmes = list(models['protectionlayer_measure'].objects.filter(protectionlayer__in=plids_chunk).get_versionedobjects(version=version))
                            objects.append(plmes)
                            m_ids = [pl.measure_id for pl in plmes]
                            if m_ids:
                                for m_ids_chunk in chunks(m_ids, 100):
                                    measures = list(models['measure'].objects.filter(**filterkwarg).filter(v_id__in=m_ids_chunk).get_versionedobjects(version=version))
                                    objects.append(measures)
                                    # attentionpoints
                                    if measures:
                                        for measures_chunk in chunks(measures, 100):
                                            objects.append(list(models['attentionpoint'].objects.filter(**filterkwarg).filter(measure__in=measures_chunk).get_versionedobjects(version=version)))

            # causetree elements
            objects.append(list(models['treeelement'].tree.filter(**filterkwarg).filter(scenario__in=sc_chunk).get_versionedobjects(version=version)))
    return objects


def make_version(sender, instance, objects, appconfig=None, **kwargs):
    """
    generic make_version, should be curried with appconfig object
    """
    if not objects:
        # only version everything when objects are not specifically included
        if instance.published:
            filterkwarg={'published_dirty': True, 'version__isnull': True}
        else:
            filterkwarg={'dirty': True, 'version__isnull': True}
        for modelname in ['suggestionlist', 'scenario', 'event', 'protectionlayer',
            'measure', 'protectionlayer_measure', 'apsuggestionlist', 'attentionpoint']:
            for item in appconfig.get_model(modelname).objects.filter(**filterkwarg):
                item.make_version(instance)
        for item in appconfig.get_model('treeelement').tree.root_nodes().filter(**filterkwarg):
            item.make_version(instance)

def restore_version(sender, instance, objects, appconfig=None, **kwargs):
    """
    generic restore_version, should be curried with appconfig object
    """
    if not objects:
        for modelname in ['suggestionlist', 'scenario', 'event', 'protectionlayer',
            'measure', 'protectionlayer_measure', 'apsuggestionlist', 'attentionpoint']:
            for item in appconfig.get_model(modelname).objects.get_newerobjects(instance):
                item.restore_version(instance)
        for item in appconfig.get_model('treeelement').tree.root_nodes().get_newerobjects(instance):
            item.restore_version(instance)
