""" code for evaluating correctness of CauseTree
"""
from analyses.deviations import models as app_models
from utils.trees import models as trees_models
from django.utils.translation import ugettext_lazy as _
from utils.versioning.models import WorkVersion

# special float functions
def fgt(float1, float2):
    """
    special comparison for floats, compensating the inaccuracy of floats
    >>> fgt(2.0, 1.0)
    True
    >>> fgt(1.0, 2.0)
    False
    >>> fgt(1.0, 1.0)
    False
    >>> fgt(1.0001, 1.0)
    False
    >>> fgt(None, 1.0)
    False
    >>> fgt(1.0, None)
    True
    >>> fgt(None, None)
    False
    """
    if not float1:
        return False
    if not float2:
        return True
    return float1 > 1.001 * float2

def feq(float1, float2):
    """
    special equality comparison for floats
    >>> feq(1.0, 1.1)
    False
    >>> feq(1.1, 1.0)
    False
    >>> feq(1.0, 1.0)
    True
    >>> feq(1.0, 1.00001)
    True
    >>> feq(None, 1.0)
    False
    >>> feq(1.0, None)
    False
    >>> feq(None, None)
    True
    """
    if not float1 and not float2:
        return True
    return not fgt(float1, float2) and not fgt(float2, float1)

def evaluate_causetree_for_deviationscenario(deviationscenario, force=False):
    """
    Base evaluation function,
    in = deviationscenario object
    out = errors

    Is not run when the scenario is not lopa_dirty. This flag gets set when the
    lopa path cache is filled and unset whenever something changes in the scenario.

    """
    if force or deviationscenario.lopa_dirty:
        clear_paths(deviationscenario)
        if deviationscenario.use_LOPA:
            tree = create_tree_for_deviationscenario(deviationscenario)
            if tree:
##                print_basetree(tree)
                (newtree, errors) = verify_causetree(tree)
##                print_tree(newtree)
                data = []
                if not errors:
                    paths = create_paths(newtree)
                    clean_dependencies_paths(paths)
##                    print_paths(paths)
                    data = calculate_paths(paths)
                    cache_paths(deviationscenario, data)
                return errors
    return []


t_EVENT = 101
t_CONDITION = 102
t_UNDETERMINED = 103
t_OR = 201
t_AND = 202

class BaseTreeElement(object):
    children = []
    item = None

class TreeElement(object):
    children = []
    item = None
    gate = t_OR
    type = t_EVENT
    errors = []

    def __repr__(self):
        if self.item:
            return repr(self.item)
        return u'Empty treeelement'

def _create_tree(items, parent_item):
    """
    Recursive function, creates basic tree based on the database data
    """
    result = []
    for item in items:
        if item.parent == parent_item:
            element = BaseTreeElement()
            element.item = item
            element.children = _create_tree(items, item)
            result.append(element)
    return result

def create_tree_for_deviationscenario(deviationscenario):
    """
    Create a tree representing the database data
    """
    version = getattr(deviationscenario, '_version', deviationscenario.version) or WorkVersion
    queryset = app_models.CauseTreeElement.tree.select_related('parent').filter(scenario=deviationscenario).get_versionedobjects(version=version).select_related_versioned()
    result = _create_tree(list(queryset), None)
    # result is a list, should only contain one element ...
    if not result:
        return None
    return result[0]

def _make_error(error, element):
    return {'error': u'%s' % error, 'element': element.item.__unicode__()}

def _resolve_undetermined_children(element, type):
    """
    recursively resolve the child items that have not yet gotten a type
    do this recursively for when multiple protection layers are nested
    """
    for child in element.children:
        if child.type == t_UNDETERMINED:
            child.type = type or element.type
            _resolve_undetermined_children(child, type)

def _verify_causetree(baseelement, skipchecks=False):
    """
    recursive evaluation of the data tree
    creates a new tree containing only events and conditions
    """
    newelement = TreeElement()
    newelement.item = baseelement.item
    errors = []
    skip = skipchecks

    ORmixerror=False

    itemtype = baseelement.item.elementtype

    # No children: check base protection layer has FF set and bail out
    if not baseelement.children:
        if itemtype == app_models.EVENT_ID:
            if baseelement.item.content_object.condition:
                newelement.type = t_CONDITION
            else:
                newelement.type = t_EVENT
            return newelement
        elif itemtype == app_models.PROTECTIONLAYER_ID:
            newelement.type = t_UNDETERMINED
            m = baseelement.item.content_object
            if not skip and not m or (m.reltype != 2 and m.reltype != 3):
                errors.append(_make_error(_('Protection layer with no children should have their FF or availability set.'), baseelement))
            newelement.errors = errors
            return newelement
        return None
    # children, but event with frequency set
    if not skip and itemtype == app_models.EVENT_ID and not baseelement.item.content_object.condition and baseelement.item.content_object.frequency:
        skip = True
    # so we have children
    # recurse through children
    newchildren = []
    for child in baseelement.children:
        e = _verify_causetree(child, skipchecks=skip)
        if e:
            newchildren.append(e)
    newelement.children = newchildren

    # determine the type
    if itemtype == app_models.EVENT_ID:
        if baseelement.item.content_object.condition:
            newelement.type = t_CONDITION
        else:
            newelement.type = t_EVENT
    elif itemtype == app_models.PROTECTIONLAYER_ID:
        newelement.type = t_UNDETERMINED

    # determine gate
    if baseelement.item.gate:
        newelement.gate = t_AND
    else:
        newelement.gate = t_OR

    # start evaluating rules
    if not skip:
        # protection layers have maximum one child
        if itemtype == app_models.PROTECTIONLAYER_ID:
            if len(newelement.children) > 1:
                errors.append(_make_error( _('Protection layer elements may only have maximum one child.'), baseelement))

        # OR gates only combine identical items
        if newelement.gate == t_OR:
            error = False
            type = None
            for child in newelement.children:
                if child.type == t_EVENT or child.type == t_CONDITION:
                    if not type:
                        type = child.type
                    else:
                        if child.type != type:
                            error=True
            if error:
                errors.append(_make_error(_('Mixing events and conditions in an OR gate is not allowed'), baseelement))
                ORmixerror=True
            # if any children are undetermined: set them here
            # do this recursively for when multiple protection layers are nested
            _resolve_undetermined_children(newelement, type)

        # strictly 1 event in AND gates leading to event
        if newelement.gate == t_AND and newelement.type == t_EVENT:
            error = False
            type = None
            eventfound = False
            for child in newelement.children:
                if child.type == t_EVENT:
                    if not eventfound:
                        eventfound = True
                    else:
                        error = True
            # if any children are undetermined: set them here
            for child in newelement.children:
                if child.type == t_UNDETERMINED:
                    if not eventfound:
                        child.type = t_EVENT
                        eventfound = True
                    else:
                        child.type = t_CONDITION
            if not eventfound:
                error = True
            if error:
                errors.append(_make_error(_('AND gates leading to an event should contain exactly one event element.'), baseelement))

        # AND gates leading to condition should only combine conditions
        if newelement.gate == t_AND and newelement.type == t_CONDITION:
            error = False
            type = None
            for child in newelement.children:
                if child.type == t_EVENT:
                    error = True
            # if any children are undetermined: set them here
            for child in newelement.children:
                if child.type == t_UNDETERMINED:
                    child.type = t_CONDITION
            if error:
                errors.append(_make_error(_('AND gates leading to a condition should contain only condition elements'), baseelement))

        # events must have event children
        if newelement.type == t_EVENT:
            if newelement.gate == t_AND:
                # we already checked that there is one event
                pass
            elif newelement.gate == t_OR and not ORmixerror:
                if newelement.children:
                    # all items are identical, check they are events
                    if newelement.children[0].type != t_EVENT:
                        errors.append(_make_error(_('Children incompatibel with event.'), baseelement))

        # conditions must have condition children
        if newelement.type == t_CONDITION:
            if newelement.gate == t_OR and not ORmixerror:
                if newelement.children:
                    # all items are identical, should be conditions
                    if newelement.children[0].type != t_CONDITION:
                        errors.append(_make_error(_('Children incompatible with condition.'), baseelement))

        newelement.errors = errors

    return newelement

def _get_errors(tree):
    errors = tree.errors[:]
    for item in tree.children:
        errors.extend(_get_errors(item))
    return errors

def verify_causetree(tree):
    newtree = _verify_causetree(tree)
    errors = _get_errors(newtree)
    return newtree, errors

class Path(object):
    """
    Path object
    list contains the path elements in reverse order
    """
    target = 0
    list = []

class PathElement(object):
    treeelement = None
    disabled = False
    freq = None
    prob = None

def copy_path(path):
    res = []
    for item in path:
        p = PathElement()
        p.treeelement = item.treeelement
        p.disabled = False
        p.freq = item.freq
        p.prob = item.prob
        res.append(p)
    return res

def create_paths(tree):
    paths = _create_paths(tree)
    result = []
    for path in paths:
        p = Path()
        p.target = path[0].treeelement.item.scenario.target_frequency
        p.list = path
        result.append(p)
    return result

def _create_path_element(element):
    el = PathElement()
    el.treeelement = element
    # determine freq and prob
    if element.type == t_EVENT:
        if element.item.elementtype == app_models.EVENT_ID:
            if element.item.content_object.frequency:
                el.freq = element.item.content_object.frequency
        elif element.item.elementtype == app_models.PROTECTIONLAYER_ID:
            if element.item.content_object.reltype == 2:
                el.freq = element.item.content_object.rel
            if element.item.content_object.reltype == 1:
                el.prob = element.item.content_object.rel
    elif element.type == t_CONDITION:
        if element.item.elementtype == app_models.EVENT_ID:
            el.prob = _calculate_probability(element)
        elif element.item.elementtype == app_models.PROTECTIONLAYER_ID:
            el.prob = _calculate_probability(element)
    return el

def _create_paths(element):
    el = _create_path_element(element)
    if element.children:
        paths = []
        if element.gate == t_OR:
            for child in element.children:
                childpaths = _create_paths(child)
                for p in childpaths:
                    t = [el]
                    t.extend(copy_path(p))
                    paths.append(t)
        elif element.gate == t_AND:
            path = [el]
            eventpaths = []
            for child in element.children:
                if child.type == t_EVENT:
                    eventpaths = _create_paths(child)
                else:
                    el2 = _create_path_element(child)
                    path.append(el2)
            for p in eventpaths:
                t = copy_path(path)
                t.extend(p)
                paths.append(t)
        return paths
    else:
        return [[el]]


def clean_dependencies_paths(paths):
    for path in paths:
        clean_dependencies_path(path)

def clean_dependencies_path(path):
    # this can be optimized by retrieving all dependencies for the tree in one query
        for item in path.list:
            item.disabled = False
        i = 0
        loop = True
        while loop and i < len(path.list):
            if not path.list[i].disabled:
                if path.list[i].treeelement.item.elementtype == app_models.PROTECTIONLAYER_ID:
                    dep = path.list[i].treeelement.item.dependencies()
                    if dep:
                        # make separate loops for causes and measures because we need to eliminate all PL's
                        # that depend on causes first
                        # first causes
                        j = i+1
                        innerloop = True
                        while innerloop and j < len(path.list):
                            # check dependencies with causes
                            if path.list[j].treeelement.item.elementtype == app_models.EVENT_ID:
                                if path.list[j].treeelement.item in dep:
                                    path.list[i].disabled = True
                            # break the searching once a freq is found, all the rest will be disabled later
                            if path.list[j].freq or path.list[i].disabled:
                                innerloop = False
                            j += 1
                        # and now protection layers
                        j = i+1
                        innerloop = True
                        while innerloop and j < len(path.list):
                            # check dependencies with protection layers
                            if path.list[j].treeelement.item.elementtype == app_models.PROTECTIONLAYER_ID:
                                if path.list[j].treeelement.item in dep:
                                    # We have a dependency
                                    # If the second one defines a frequency, we need to keep it
                                    if path.list[j].freq:
                                        path.list[i].disabled = True
                                    # if one of the PL's is already disabled: do nothing
                                    elif path.list[i].disabled or path.list[j].disabled:
                                        pass
                                    else:
                                        if path.list[i].prob:
                                            if path.list[j].prob:
                                                if path.list[i].prob < path.list[j].prob:
                                                    path.list[j].disabled = True
                                                else:
                                                    path.list[i].disabled = True
                                            else:
                                                path.list[j].disabled = True
                            # break the searching once a freq is found, all the rest will be disabled later
                            if (path.list[j].freq and not path.list[j].treeelement.item.elementtype == app_models.PROTECTIONLAYER_ID) or path.list[i].disabled:
                                innerloop = False
                            j += 1

            if path.list[i].freq:
                loop = False
            i += 1
        while i < len(path.list):
            path.list[i].disabled = True
            i += 1

def calculate_paths(paths):
    result = []
    i = 0
    for path in paths:
        p = {'initiating': path.list[-1].treeelement.item, 'final': path.list[0].treeelement.item}
        p['frequency'] = _calculate_path(path.list)
        p['target'] = path.target
        p['items'] = path.list
        p['path_id']= i
        result.append(p)
        i += 1
    return result

def _calculate_path(path):
    if path:
        if path[0].disabled:
            return _calculate_path(path[1:])
        else:
            if path[0].prob:
                return path[0].prob * _calculate_path(path[1:])
            elif path[0].freq:
                return path[0].freq
            else:
                return _calculate_path(path[1:])
    else:
        return 1

def _calculate_probability(element):
    if element.item.elementtype == app_models.EVENT_ID:
        if element.item.content_object.probability:
            return element.item.content_object.probability
        else:
            if element.children:
                if element.gate == t_OR:
                    p = 0
                    for child in element.children:
                        p += _calculate_probability(child)
                    return p
                elif element.gate == t_AND:
                    p = 1
                    for child in element.children:
                        p = p * _calculate_probability(child)
                    return p
            return 0
    elif element.item.elementtype == app_models.PROTECTIONLAYER_ID:
        if element.item.content_object.reltype == 3:
            if element.item.content_object.rel:
                return 1 - element.item.content_object.rel
            return 0
        elif element.children:
            p = 0
            for child in element.children:
                p += _calculate_probability(child)
            return p
    return 0


def clear_paths(scenario):
    # delete existing paths
    ps = app_models.DeviationLopaPath.objects.filter(scenario=scenario)
    for path in ps:
        app_models.DeviationLopaPathElement.objects.filter(path=path).delete()
        path.delete()

def cache_paths(scenario, paths):
    # cache the paths
    i = 0
    for path in paths:
        p = app_models.DeviationLopaPath(
            scenario=scenario,
            number = i,
            initiating = u'%s' % path['initiating'],
            final = u'%s' % path['final'],
            frequency = path['frequency'],
            target = path['target'],
            high = fgt(path['frequency'], path['target'])
        )
        p.save()
        j = 0
        for element in path['items']:
            e = app_models.DeviationLopaPathElement(
                path = p,
                number = j,
                name = u'%s' % element.treeelement.item,
                prob = element.prob,
                freq = element.freq,
                disabled = element.disabled,
                elementtype = element.treeelement.item.get_elementtype_for_tree()
            )
            e.save()
            j += 1
        i += 1
    scenario.set_lopa_dirty(flag=False)

def print_basetree(tree, prefix=''):
    print prefix, tree.item
    for item in tree.children:
        print_basetree(item, prefix + '  ')

def print_tree(tree, prefix=''):
    print prefix, tree.item, tree.type, tree.gate
    for item in tree.children:
        print_tree(item, prefix + '  ')

def print_paths(paths):
    for path in paths:
        print 'Path target: ', path.target
        for element in path.list:
            print '  ', element.treeelement.item, element.treeelement.type, element.treeelement.gate, element.freq, element.prob, element.disabled

