"""
Tests for news app
"""

import datetime

from django import test
from django.conf import settings
from django.contrib.auth.models import User, Permission, AnonymousUser
from django.contrib.contenttypes.models import ContentType

from utils.news import models as app_models
from utils.versioning.models import WorkVersion
from utils import tests as utils_tests

class NewsTestCase(test.TestCase):
    fixtures = ['dummyusers', 'news']

    def setUp(self):
        # give the users permissions

        # authorisations for users:
        # 1 = authorised to edit entries
        # 2 = only authorised to display
        # 3 = unauthorised to even display

        u = User.objects.get(pk=1)

        permissions = [
            ('display_news','news','entry'),
            ('edit_news','news','entry'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=2)
        permissions = [
            ('display_news','news','entry'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

    def test_entry_views(self):
        base_url = '/news/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3,)

        class entries(utils_tests.tablelist_TestDefinition):
            description = 'list of news entries'
            url = base_url
            users = [1,2,]
            rowcount = 2
            contains = [{'url': base_url + 'add/', 'users': [1,]},]

        class entry(utils_tests.TestDefinition):
            description = 'display a news entry'
            url = base_url + '1/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': [1,]},
            ]

        class entry_edit(utils_tests.edit_TestDefinition):
            description = 'edit a news entry'
            url = base_url + '1/edit/'
            users = [1,]
            data = {
                "title": "Great news!",
                "body": "Let's dance.",
                "management-referer": base_url + '1/'
            }
            redirect_url = base_url + '1/'

        class entry_delete(utils_tests.delete_TestDefinition):
            description = 'delete a news entry'
            users = [1,]
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Entry(title="Oh my", body="hmm")
                newitem.save(defaultuser)
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Entry.objects.get(id=self.pk).delete()

        tests = [
            entries(),
            entry(),
            entry_edit(),
            entry_delete(),
        ]

        utils_tests.run_tests(self, users, tests)


    def test_templatetags(self):
        from django.template import Context, Template
        d = {'STATIC_URL': 'x', 'user': User.objects.get(pk=1)}
        t = Template("""
            {% load news %}
            {% get_entries 5 as entries %}
        """)
        response = t.render(Context(d))

class NewsSearchTestCase(NewsTestCase):
    def test_searches(self):
        users = (0,1,2,3,)

        defaultuser = User.objects.get(pk=1)

        class base_search(utils_tests.TestDefinition):
            description = ''
            searchclass = None
            searchitems = []
            users = [1,2,]
            def runtests(self, testcase, allowed, contains, notcontains):
                user = testcase.user or AnonymousUser()
                for item in self.searchitems:
                    res = allowed and item['result'] or 0
                    testcase.assertEquals(len(self.searchclass().search(item['query'], user, WorkVersion, None)), res)

        class search_entry(base_search):
            description = 'search a news entry'
            searchclass = app_models.News_EntrySearch
            searchitems = [
                { "query": 'planop', "result": 2},
                { "query": 'xxx', "result": 0},
            ]

        tests = [
            search_entry(self),
        ]

        utils_tests.run_tests(self, users, tests)
