from django.utils.encoding import smart_unicode, force_unicode
from django.utils.html import escape, conditional_escape
from django.utils.html import linebreaks
from django.utils.safestring import mark_safe
from django.utils.dateformat import format, time_format
from django.conf import settings
import datetime
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _

class SheetField(object):
    """Represents a single field of a table.

    ``label`` defines a display name for this field used for output.

    ``name`` is the internal name of the field. Normally you don't need to
    specify this, as the attribute that you make the field available under
    is used. However, in certain circumstances it can be useful to override
    this default, e.g. if you want a field to not
    use the model field name.

    ``default`` is the default value for this field. If the data source
    does provide ``None`` for a row, the default will be used instead. Note
    that whether this effects ordering might depend on the table type (model
    or normal). Also, you can specify a callable, which will be passed a
    ``BoundRow`` instance and is expected to return the default to be used.

    Additionally, you may specify ``data``. It works very much like
    ``default``, except it's effect does not depend on the actual cell
    value. When given a function, it will always be called with a row object,
    expected to return the cell value. If given a string, that name will be
    used to read the data from the source (instead of the column's name).

    Note the interaction with ``default``. If ``default`` is specified as
    well, it will be used whenver ``data`` yields in a None value.

    """

    # Tracks each time a Field instance is created. Used to retain order.
    creation_counter = 0

    def __init__(self, label=None, name=None, default=u'', data=None,
                 link=None, auto_id=True):
        self.label = label
        self.name = name
        self.default = default
        self.data = data
        self.link = link
        self.auto_id = auto_id

        self.creation_counter = SheetField.creation_counter
        SheetField.creation_counter += 1

    def render_field(self, instance):
        field = self.get_field(instance)
        id = self.get_id(instance)
        if id:
            return u'<td><div id="%(id)s">%(field)s</div></td>' % {'field': field, 'id': id}
        return u'<td>%(field)s</td>' % {'field': field}

    def render_label(self, instance):
        label = self.get_label(instance)
        return u'<th>%(label)s</th>' % {'label': label}

    def render(self, sheet, instance):
        if sheet.omit_empty:
            if not self.get_data(instance):
                return u''
        field = self.render_field(instance)
        label = self.render_label(instance)
        return u'<tr>%(label)s%(field)s</tr>' % {'field': field, 'label': label}

    def get_data(self, instance):
        value=None
        if callable(self.data):
            value = self.data(instance)
        elif isinstance(self.data, basestring):
            try:
                value = getattr(instance, self.data)
            except:
                try:
                    value = instance.get(self.data)
                except:
                    pass
        if value is None:
            value = self.default
        return value

    def get_compare_data(self, instance):
        return self.get_data(instance)

    def get_field(self, instance):
        value = self.get_data(instance)
        url = self.get_link(value, instance)
        value = self.format_value(value)
        if url:
            return u'<a href="%s">%s</a>' % (url, value)
        else:
            return value

    def get_label(self, instance):
        if self.label:
            label = smart_unicode(self.label)
            if label[-1] not in ':?.!':
                label += ':'
        else:
            label = u'&nbsp;'
        return label

    def get_id(self, instance):
        if self.auto_id:
            return u'id_%s' % self.name
        return u''

    def format_value(self, value):
        return mark_safe(u'%s' % conditional_escape(force_unicode(value)))

    def get_link(self, value, instance):
        if self.link:
            if callable(self.link):
                return self.link(instance)
            else:
                return self.link
        return None

class EmptyField(SheetField):
    """
    SheetField that always displays as an empty value
    """
    def get_data(self, instance):
        return None

    def get_field(self, instance):
        return mark_safe(u'&nbsp;')

class TextField(SheetField):
    """
    SheetField for which the linebreaks filter is used on the value.
    (this surrounds it with <p></p> tags and converts \n to <br />)
    """
    def format_value(self, value):
        return mark_safe(u'%s' % linebreaks(conditional_escape(force_unicode(value))))

class _DTField(SheetField):
    def format_value(self, value):
        try:
            return format(value, self.format)
        except:
            return u''

class DateField(_DTField):
    """
    display a date, accepts a format argument, falling back on settings.DATE_FORMAT
    """
    def __init__(self, format=None, *args, **kwargs):
        super(DateField, self).__init__(*args, **kwargs)
        self.format = format or settings.DATE_FORMAT

class DateTimeField(_DTField):
    """
    display a datetime, accepts a format argument, falling back on settings.DATETIME_FORMAT
    """
    def __init__(self, format=None, *args, **kwargs):
        super(DateTimeField, self).__init__(*args, **kwargs)
        self.format = format or settings.DATETIME_FORMAT

class TimeField(_DTField):
    """
    display a time, accepts a format argument, falling back on settings.TIME_FORMAT
    """
    def __init__(self, format=None, *args, **kwargs):
        super(TimeField, self).__init__(*args, **kwargs)
        self.format = format or settings.TIME_FORMAT

    def format_value(self, value):
        try:
            return time_format(value, self.format)
        except:
            return u''

class BooleanField(SheetField):
    """
    display a boolean as Yes or No
    """
    def format_value(self, value):
        if value:
            return _('Yes')
        else:
            return _('No')

class ChoicesField(SheetField):
    """
    display a value like if it was selected from a choices list of tuples (eg a choicefield)
    the label fromm the choices is used to display the value
    """
    def __init__(self, choices=None, *args, **kwargs):
        super(ChoicesField, self).__init__(*args, **kwargs)
        self.choices = choices

    def format_value(self, value):
        str_value = force_unicode(value)
        if self.choices:
            for option_value, option_label in self.choices:
                option_value = force_unicode(option_value)
                if option_value == str_value:
                    value = option_label
        return mark_safe(u'%s' % conditional_escape(force_unicode(value)))

class URLField(SheetField):
    """
    display a url, if needed adds 'http://'
    accepts an add_link argument
    """
    def __init__(self, add_link=False, *args, **kwargs):
        super(URLField, self).__init__(*args, **kwargs)
        self.add_link = add_link

    def _normalize(self,value):
# TODO: better ways to normalize this
        if value and value[:4] != 'http':
            return "http://%s" % value
        return value

    def format_value(self, value):
        return mark_safe(u'%s' % conditional_escape(force_unicode(self._normalize(value))))

    def get_link(self, value, instance):
        if self.add_link:
            return self._normalize(value)
        return None

class ForeignKeyField(SheetField):
    """
    display a foreign key field (uses the __unicode__ representation of the foreign object)
    accepts an add_link argument, which calls get_absolute_url() on the foreign object
    """
    def __init__(self, add_link=False, *args, **kwargs):
        super(ForeignKeyField, self).__init__(*args, **kwargs)
        self.add_link = add_link

    def get_link(self, value, instance):
        if self.add_link:
            if hasattr(value, "get_absolute_url"):
                return value.get_absolute_url()
        return None

class ManyToManyField(SheetField):
    """
    display the objects from a manytomany relationship (uses the __unicode__ representation of the foreign objects)
    accepts an add_link argument, which calls get_absolute_url() to add a link on each of the foreign objects
    the objects are separated with a ',' unless the newlines argument is used in which case a <br /> is used in between
    """
    def __init__(self, add_link=False, newlines=False, *args, **kwargs):
        super(ManyToManyField, self).__init__(*args, **kwargs)
        self.add_link = add_link
        self.newlines = newlines

    def get_data(self, instance):
        return getattr(instance, self.data).all()

    def get_field(self, instance):
        object_list = self.get_data(instance)
        if not object_list:
            return self.default
        else:
            output = []
            for obj in object_list:
                value = self.format_value(obj)
                url = self.get_link(obj)
                if url:
                    output.append(u'<a href="%s">%s</a>' % (url, value))
                else:
                    output.append(value)
            if self.newlines:
                return u'<br />'.join(output)
            return u', '.join(output)

    def get_link(self, object):
        if self.add_link:
            if hasattr(object, "get_absolute_url"):
                return object.get_absolute_url()
        return None

class FileField(SheetField):
    """
    display a filefield, uses the url attribute of the file for the link
    """
    def get_link(self, value, instance):
        return value.url

def formfield_callback(field, add_link=False):
    from django.db import models
    label = capfirst(field.verbose_name)
    mapping = [
        {'fieldtype': models.URLField, 'sheetfieldtype': URLField},
        {'fieldtype': models.DateTimeField, 'sheetfieldtype': DateTimeField},
        {'fieldtype': models.DateField, 'sheetfieldtype': DateField},
        {'fieldtype': models.TimeField, 'sheetfieldtype': TimeField},
        {'fieldtype': models.ForeignKey, 'sheetfieldtype': ForeignKeyField, 'kwargs': {'add_link': add_link}},
        {'fieldtype': models.ManyToManyField, 'sheetfieldtype': ManyToManyField, 'kwargs': {'add_link': add_link}},
        {'fieldtype': models.BooleanField, 'sheetfieldtype': BooleanField},
        {'fieldtype': models.TextField, 'sheetfieldtype': TextField},
        {'fieldtype': models.FileField, 'sheetfieldtype': FileField},
        ]
    if isinstance(field, models.AutoField):
        return None
    if field.choices:
        return ChoicesField(data=field.name, label=label, choices=field.get_flatchoices())
    for map in mapping:
        if isinstance(field, map['fieldtype']):
            return map['sheetfieldtype'](data = field.name, name = field.name, label = label, **map.get('kwargs',{}))
    return SheetField(data=field.name, name=field.name, label=label)
